#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <KernelKit.h>

typedef struct Track0
{
	long VersionNumber;
	long FormatDate;
	long FirstBitMapSector;
	long BitMapSize;
	long FirstDirSector;
	long TotalSector;
	long BytePerSector;
	long DirBlockHint;
	long FreeSectorHint;
	long SectorUsed;
	char Removable;
	char Fillerc[3];
	char VolName[32];
	long clean_shutdown;
	long root_ref;
	long root_modbits;
	long Filler[107];
} Track0;

typedef struct DirectoryBlockHeader
{
	long NextDirectoryBlock;
	long LinkUp;
	long Filler[14];
} DirectoryBlockHeader;

typedef struct FileEntry
{
	char FileName[64];
	long FirstAllocList;	// block (*512) of directory
	long LastAllocList;		// block (if FileType=-1 dir)
	long FileType;
	long CreationDate;
	long ModificationDate;
	long LogicalSize;
	long PhysicalSize;
	long Creator;
	long rec_id;
	long modbits;
	long unsiged2;
	long unsiged3;
	long unsiged4;
	long unsiged5;
	long unsiged6;
	long unsiged7;
} file_entry;

typedef struct DirectoryBlock
{
	DirectoryBlockHeader Header;
	FileEntry Entries[63];
	long filler[16];
};

int RestoreFatFile(int fd,char *fname,long loc,long s)
{
	ulong fat[128];
	char buffer[512];
	printf("Restore '%s', loc=%d,size=%d\n",fname,loc,s);
	lseek(fd,loc*512,SEEK_SET);
	read(fd,(void *)fat,512);
	if (fat[0])
	{
		fprintf(stderr,"Error: '%s' has more than one fat table.\n");
		return 0;
	}
	if (FILE *FOut=fopen(fname,"wb"))
	{
		int fatindex=2;
		while (fat[fatindex]!=-1 && s>0)
		{
			lseek(fd,fat[fatindex]*512,SEEK_SET);
			for (int i=0;(i<fat[fatindex+1]) && (s>0);i++)
			{
				int red;
//				printf("to go: %d\n",s);
				red=read(fd,buffer,512);
				fwrite(buffer,(s<512)?s:red,1,FOut);
				s-=red;
			}
			fatindex+=2;
		}
		fclose(FOut);
		return 1;
	}
	else
	{
		fprintf(stderr,"Couldn't open '%s'\n",fname);
		return 0;
	}
}
void RestoreSectorFile(int fd,char *fname, long loc,long s)
{
	char buffer[512];
	
	printf("Restore '%s', loc=%d,size=%d\n",fname,loc,s);
	if (FILE *FOut=fopen(fname,"wb"))
	{
		lseek(fd,loc*512,SEEK_SET);
		while (s>0)
		{
			int red;
		//	printf("to go: %d\n",s);
			red=read(fd,buffer,512);
			fwrite(buffer,(s<512)?s:red,1,FOut);
			s-=red;
		}
		fclose(FOut);
	}
}

int readonly=1;

int depth=0;
void ReadDir(int fd,long loc,char *srcdir,char *destdir)
{
	DirectoryBlock d;
	int i;
	int loopcheck=0;
	
	depth++;
	while ((loc>0) && (loopcheck++<10))
	{
		lseek(fd,loc*512,SEEK_SET);
		read(fd,(void *)&d,sizeof(d));
		for (i=0;i<64;i++)
		{
			int entlen=strlen(d.Entries[i].FileName);
			if (!entlen) 
				continue;
			char newsrc[120];
			char newdst[120];
			if (srcdir[0])
			{
				if (srcdir[entlen]=='/')
					strcpy(newsrc,srcdir+entlen+1);
				else
					strcpy(newsrc,srcdir+entlen);
			}
			else
				newsrc[0]=0;
			sprintf(newdst,"%s/%s",destdir,d.Entries[i].FileName);
//			printf("orig='%s','%s', new='%s','%s'\n",
//				srcdir,destdir,newsrc,newdst);
//			for (int j=1;j<depth;j++) printf("\t");
			switch(d.Entries[i].FileType)
			{
				case -1:
					if (readonly && !srcdir[0])
					printf("drw-rw-rw-   1 wildcard design   %8d %s\n",// %d\n",
						d.Entries[i].PhysicalSize,
						d.Entries[i].FileName);//,
						//d.Entries[i].FirstAllocList);
					// recurse only if 
					// 1) the source dir is blank, and we're copying or...
					// 2) the file/dir name matches exactly
					if ((!srcdir[0] && !readonly) || 
						(strncmp(d.Entries[i].FileName,srcdir,entlen)==0
					 && (srcdir[entlen]==0 || srcdir[entlen]=='/')))
					{
						if (!readonly)
						{
							char cmd[200];
							sprintf(cmd,"mkdir %s >/dev/null 2>/dev/null",newdst);
							//printf("%s\n",cmd);
							system(cmd);
						}
						ReadDir(fd,d.Entries[i].FirstAllocList,newsrc,newdst);
					}
					break;
				default:
					if (readonly && (!srcdir[0] || strcmp(srcdir,d.Entries[i].FileName)==0))
					printf("%crw-rw-rw-   1 wildcard design   %8d %s\n", // %d\n",
						'-',
						//(d.Entries[i].FirstAllocList&0x80000000)?'s':'f',
						d.Entries[i].LogicalSize,
						d.Entries[i].FileName);//,
					//	d.Entries[i].FirstAllocList&0x7fffffff);
					if (!readonly && (!srcdir[0] || strcmp(srcdir,d.Entries[i].FileName)==0))
					{
					//	printf("Get file %s\n",newdst);
						if (d.Entries[i].FirstAllocList&0x80000000)
							RestoreSectorFile(fd,newdst,
								d.Entries[i].FirstAllocList&0x7fffffff,
								d.Entries[i].LogicalSize);
						else
							RestoreFatFile(fd,newdst,
								d.Entries[i].FirstAllocList&0x7fffffff,
								d.Entries[i].LogicalSize);
					}
			}
		}
		loc=d.Header.NextDirectoryBlock;
//printf("next=%d\n",loc);
	}
	depth--;
}

main(int argc,char *argv[])
{
	int fd;
	Track0 t;
	DirectoryBlock d;
	long loc;
	int i;
	char destdir[120];
	char srcdir[120];
	
	if (argc<3)
	{
		fprintf(stderr,"Usage: beread <device filename> <directory> [<dest directory>]\n");
		return 0;
	}
	if (argc==4)
	{	
		strcpy(destdir,argv[3]);
		readonly=0;
	}
	else
	{
		destdir[0]=0;
	}
	if (argv[2][0]=='/')
		strcpy(srcdir,argv[2]+1);
	else
		strcpy(srcdir,argv[2]);
	if ((fd=open(argv[1],O_RDONLY))==-1)
	{
		fprintf(stderr,"oops.. couldn't open.\n");
		return 0;
	}
	read(fd,(void *)&t,sizeof(t));
	printf("Disk Label: %s\n",t.VolName);
//	printf("first dir sector=%d\n",t.FirstDirSector);
	ReadDir(fd,t.FirstDirSector,srcdir,destdir);

	close(fd);
	return 0;
}